-- Timing subsytem
-- SNS LLRF Digital board
-- Target: XC2S150
-- Larry Doolittle  <LRDoolittle@lbl.gov>
-- January 18-20, 2002
-- Converted to VHDL February 28, 2002

-- wishlist:
--   record length of RF pulse, for readback by software
--   double buffer ttable? easy with an additional address bit


-- Input "ck" is the 40 MHz acquisition clock.
--
-- quad_enable is the 10 MHz 1-out-of-4 synchronous enable signal
--
-- rf is the rfon input, assumed already synchronized to 40 MHz
--
-- halt and trig_fallback are control bits set by the host
--
-- write_pulse, host_* are the host interface, whereby the host
-- can read and write the 8 16 bit timing control registers.
-- write_pulse is assumed long compared to the 100 ns cycle,
-- and multiple writes to the register file are OK.

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;


entity rf_timer is
  port (
	ck:              in std_logic;  -- 40 MHz
	quad_enable:     in std_logic;  -- 10 MHz
	halt:            in std_logic;
	rf:              in std_logic;
	self_retrigger:  in std_logic;
	ignore_rf_off:   in std_logic;
	trig_fallback:   in std_logic;
	write_pulse:     in std_logic;
	host_address:    in std_logic_vector(2 downto 0);
	host_data:       in std_logic_vector(15 downto 0);
	host_read:       out std_logic_vector(15 downto 0);
	feedback_run:    out std_logic;
	integrate_run:   out std_logic;
	feedforward_start: out std_logic;
	trace_enable:    out std_logic;
	adc_power:       out std_logic;
	host_interrupt:  out std_logic;  -- happens at end of every pulse
	rf_gate:         out std_logic;
	trigger_skipped: out std_logic  -- rf edge when we didn't want it
  );
end rf_timer;

architecture virtex of rf_timer is

-- Synchronous 16 x 16 RAM, constructed of 16 Virtex LE's
component RAM16X16S is
    port(
        A0:    in  std_logic;
        A1:    in  std_logic;
        A2:    in  std_logic;
        A3:    in  std_logic;
	D:     in  std_logic_vector(15 downto 0);
	O:     out std_logic_vector(15 downto 0);
	WCLK:  in  std_logic;
	WE:    in  std_logic
    );
end component;

-- Give ourselves oodles of time for this logic, and make everything
-- 10 MHz synchronous.
signal phase1:  std_logic;
signal phase2:  std_logic;
signal phase3:  std_logic;
signal a_select: std_logic;

-- States:                       state_advance:     if (state_advance) reload:
--   0  8.437ms  Wait for trigger         rf                            t2
--   1  1.434ms  Idle                          t2==0                    t1
--   2  0.001ms  Idle                            1                      t2
--   3  0.030ms  ADC warmup                    t2==0                    t2 
--   4  0.020ms  Feedforward only        !rf | t2==0                    t2
--   5  1.015ms  Feedback                !rf | t2==0                    t2
--   6  0.030ms  Acquire decay curve           t2==0
--   7  5.700ms  Inhibit retrigger             t1==0                    t1

-- But: if ignore_rf_off is set, states 4 and 5 continue
-- until timer expiration even when rf is false.
-- And if self_retrigger is set, advance out of state 0 if t1==0.
--
-- In a live system, both those command flags would be cleared.
-- If you only have a short trigger pulse, that is low by the time
-- rf should be turned on, set ignore_rf_off.  For triggerless bench
-- testing, set both ignore_rf_off and self_retrigger.

-- The state register is used so much, I choose a single letter for
-- its identifier: "s".  This keeps the source lines shorter.
-- Of course, the rewrite from Verilog to VHDL makes this a losing battle.
-- At least, since s is three bits wide, VHDL lets me use octal constants,
-- instead of writing out "101" for O"5".  Not quite as good as Verilog's 5.
signal s: std_logic_vector(2 downto 0);
signal t1: std_logic_vector(17 downto 0);
signal t2: std_logic_vector(15 downto 0);
signal old_rf: std_logic;

signal timer_zero:      std_logic;
signal timer_inhibit:   std_logic;
signal rf_kill_advance: std_logic;
signal start_cycle:     std_logic;
signal state_advance:   std_logic;
signal ext_trigger:     std_logic;

signal t1_reload: std_logic;
signal t2_reload: std_logic;
signal ta: std_logic_vector(3 downto 0);
signal ttable: std_logic_vector(15 downto 0);   -- output of RAM
signal table_write: std_logic;
signal s0, s1, s2, s3, s4, s5, s6, s7: std_logic;
signal t1_zero, t2_zero: std_logic;

-- on to the assignments and processes
begin

-- such garbage
s0 <= '1' when (s=O"0") else '0';
s1 <= '1' when (s=O"1") else '0';
s2 <= '1' when (s=O"2") else '0';
s3 <= '1' when (s=O"3") else '0';
s4 <= '1' when (s=O"4") else '0';
s5 <= '1' when (s=O"5") else '0';
s6 <= '1' when (s=O"6") else '0';
s7 <= '1' when (s=O"7") else '0';
t1_zero <= '1' when (t1="0000000000000000") else '0';
t2_zero <= '1' when (t2="000000000000000000") else '0';


timer_zero      <= ((s0 or s7) and t1_zero) or
                    (not(s0 or s7) and t2_zero);
timer_inhibit   <= (s0) and not self_retrigger;
rf_kill_advance <= (s4 or s5) and not rf and not ignore_rf_off;
ext_trigger     <= rf and not old_rf;
start_cycle     <= (s0) and ext_trigger;
state_advance   <= start_cycle or (timer_zero and not timer_inhibit) or
                   (s2) or rf_kill_advance;

phase_stuff: process (ck) begin
    if (ck'event and ck = '1') then
	phase1 <= quad_enable;
	phase2 <= phase1;
	phase3 <= phase2;
	-- of course, because all these quantities are registered,
	-- a_select ends up high during phase2 and phase3.
	a_select <= phase1 or phase2;
    end if;
end process phase_stuff;

-- It's OK to register all of these, and delay everything 100 ns.
-- That amount of time shift for any one signal is probably irrelevant,
-- and registering them all balances the delays and guarantees no glitches.
function_select: process (ck) begin
    if (ck'event and ck='1') then
      if (phase1='1') then
	rf_gate          <= ( s4 or s5 ) and (rf or ignore_rf_off);
	adc_power        <= s3 or s4 or s5 or s6;
	feedback_run     <= s5;
	integrate_run    <= s4 or s5;
	feedforward_start <= s3 and t2_zero;
	trace_enable     <= s4 or s5 or s6;
	host_interrupt   <= s6 and t2_zero;
      end if;
    end if;
end process function_select;

t1_reload <= halt or (state_advance and (s1 or s7));
t2_reload <= state_advance and (s0 or s2 or s3 or s4 or s5);

-- The RAM address is multiplexed between host R/W, and
-- making the output available for time reloading.
address_mux: process (a_select, host_address, s) begin
	if (a_select = '1') then
		ta <= "0" & host_address;
	else
		ta <= "0" & s;
	end if;
end process address_mux;

table_write_setup: process (ck) begin
    if (ck'event and ck='1') then
    	-- only active during phase3
	table_write <= phase2 and write_pulse;
    end if;
end process table_write_setup;

timer_table: RAM16X16S port map (
	A0    => ta(0),
	A1    => ta(1),
	A2    => ta(2),
	A3    => ta(3),
	D     => host_data(15 downto 0),
	O     => ttable(15 downto 0),
	WCLK  => ck,
	WE    => table_write
);
--ttable <= "0000000000000000";

timer_advance: process(ck) begin
    if (ck'event and ck='1') then
      if (phase1='1') then
	old_rf <= rf;
	trigger_skipped <= not s0 and ext_trigger;
	if (halt='1') then
		s <= O"0";
	elsif (state_advance='1') then
		s <= s+1;
	end if;
	if (t1_reload='1') then
		t1 <= ttable & "00";  -- bit concatenation
	else
		t1 <= t1-1;
	end if;
	if (t2_reload='1') then
		t2 <= ttable;
	else
		t2 <= t2-1; 
	end if;
      end if;
      if (phase3='1') then
        host_read <= ttable;
      end if;
    end if;
end process timer_advance;

end virtex;
