/*
 * mmap 64K from 0x40000000 through 0x4000ffff
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/poll.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stddef.h>    /* offsetof() */
#include <string.h>    /* strtok() */
#include <stdlib.h>    /* strtol() */

#include "llrf.h"

struct fpga_state {
#define  VAR(t,n,v) t v;
#define IVAR(t,n,v,p,r) t v;
#define  ARRAY_VAR(t,n,v,l) t v[l];
#define ARRAY_IVAR(t,n,v,l,p,r) t v[l];
#include "llrf-state.h"
#undef VAR
#undef IVAR
#undef ARRAY_VAR
#undef ARRAY_IVAR
} fs;

struct xfer_element{
	int fpga_o;
	int n;
	size_t o;
} int_xfer_list[] = {
	{ 0x0004,    1, offsetof(struct fpga_state, firmware_version)},
	{ 0x0008,    1, offsetof(struct fpga_state, status)},
	{ 0x000c,    1, offsetof(struct fpga_state, errors)},
	{ 0x0014,    1, offsetof(struct fpga_state, pulse_length)},
};

inline char *raw_utoa(unsigned int i, char *p)
{
	do {
		*p-- = '0' + i % 10;
		i /= 10;
	} while (i > 0);
	return p;
}

/* Note: this _does_ get the right answer for i == INT_MIN,
 * on the limited number of machines I have available.
 * I haven't tested it on a ones-complement machine.
 * See March 2000 thread in comp.unix.programmer, with
 * subject "undefined itoa in shared library" */
inline char *raw_itoa(int i, char *p)
{
	unsigned int x = i>=0 ? i : -i;
	p = raw_utoa(x,p);
	if (i<0) *p-- = '-';
	return p;
}

void import_raw_trace(int addr, int last_address, short dx[]) {
	int i, j;
	for (i=0, j=last_address%1024+1; i<1024; i++, j++) {
		if (j>=1024) j-=1024;
		dx[i] = (inw(addr+j*4) & 0xfff) - 2048;
	}
}

void import_decimated_trace(int addr, int decimate, int last_address, short dr[], short di[])
{
	/* only works if decimate+1 is a power of two */
	int i, j, k;
	k = last_address%(1024*(decimate+1));
	k = 255-k/(4*(decimate+1));
	for (i=0,j=0; i<256; i++, k++) {
#if 0
		if (i==0) printf("addr 0x%x: %d-%d %d-%d\n",addr,
		  (inw(addr+(j+0)*4)&0xfff), (inw(addr+(j+2)*4)&0xfff),
		  (inw(addr+(j+1)*4)&0xfff), (inw(addr+(j+3)*4)&0xfff));
#endif
		if (k>=256) k-=256;
		dr[k] = ((inw(addr+(j+0)*4)&0xfff) - (inw(addr+(j+2)*4)&0xfff))/2;
		di[k] = ((inw(addr+(j+1)*4)&0xfff) - (inw(addr+(j+3)*4)&0xfff))/2;
		j += 4*(decimate+1);
		if (j>=1024) j-=1020;
	}
}

int analyze_decay(int n, short d[], double r[]);
void xfer_from_fpga(int nintr)
{
	int i, j, t;
	double decay_result[2];
	for (i=0;i<sizeof(int_xfer_list)/sizeof(*int_xfer_list);i++) {
		for (j=0;j<int_xfer_list[i].n;j++) {
			*((int *)((char *)&fs+int_xfer_list[i].o)+j) =
				inw(int_xfer_list[i].fpga_o+j);
		}
	}
#if 0
	printf("# done with int_xfer_list\n");
#endif

	import_decimated_trace(0x8000, fs.trace_decimate, fs.pulse_length, fs.trace_ar, fs.trace_ai);
	import_decimated_trace(0xa000, fs.trace_decimate, fs.pulse_length, fs.trace_br, fs.trace_bi);
	import_decimated_trace(0xc000, fs.trace_decimate, fs.pulse_length, fs.trace_cr, fs.trace_ci);
	import_raw_trace      (0x4000,                    fs.pulse_length, fs.trace_x);
#if 0
	printf("# done with trace input\n");
#endif

	/* extra ADC inputs are not (quite) simple register transfers */
	i=fs.cycle_count&7;
/* test mode can wait forever for the "complete" bit */
#ifdef __arm__
	fs.extra_adc[i  ]=get_adc(i  );
	fs.extra_adc[i+8]=get_adc(i+8);
#endif
	putchar('.'); fflush(stdout);

	/* ====== now we are done touching the FPGA, on to post-processing ====== */

	fs.nintr = nintr;   /* number of interrupts read from pulse device */

	/* update error and cycle counters */
	/* printf("# update error and cycle counters\n"); */
	t = fs.errors;
	for (i=0; i<8; i++) {
		if (t&1) fs.error_count[i]++;
		t=t>>1;
	}
	fs.cycle_count++;

	/* analyze the decay waveform */
	/* printf("# analyze decay\n"); */

	{
		int o, l;
		/* timer[5] is t_decay */
		o = 1024-fs.timer[5]*4+fs.decay_ignore;
		if (o<0) o=0;
		l = fs.decay_length;
		if (o+l>1024) l=1024-o;
		/* printf("o=%d l=%d\n",o,l); */
		fs.decay_stat = analyze_decay(l,fs.trace_x+o,decay_result);
	}
	if (fs.decay_stat == 0) {
		fs.delta_f   = decay_result[0];
		fs.bandwidth = decay_result[1];
	}
}

int set_int   (int i, char *line);
int set_double(int i, char *line);
void fpga_set_int(size_t o, int ix, int reg);
void fpga_set_dac(size_t o, int ix, int reg);
void fpga_set_setpoint(size_t o, int ix, int reg);
void fpga_set_gain(size_t o, int ix, int reg);

const struct input_cmd {
	char c[2];
	int (*func)(int i, char *line);
	size_t o;
	void (*hwrite)(size_t o, int ix, int reg);
	int r;
} input_cmds[] = {
#define IVAR(t,n,v,p,r) { n, set_##t, offsetof(struct fpga_state, v), p, r },
#define ARRAY_IVAR(t,n,v,l,p,r) { n, set_##t, offsetof(struct fpga_state, v[0]), p, r },
#define VAR(t,n,v)
#define ARRAY_VAR(t,n,v,l)
#include "llrf-state.h"
#undef VAR
#undef IVAR
#undef ARRAY_VAR
#undef ARRAY_IVAR
};

int set_int(int i, char *line)
{
	char *s, *s2;
	int val;
	size_t o;
	int o2;
	o = input_cmds[i].o;
	s=strtok(NULL," ");
	if (s==NULL) return 1;
	o2 = strtol(s,&s2,0);
	if (*s2 != '\0') return 1;  /* offset must not have invalid characters */
	s=strtok(NULL," ");
	val = strtol(s,&s2,0);
	if (s==s2) return 1;     /* no rval */
	printf("# writing %d to offset 0x%x index %d\n",val,o,o2);
	*((int *)((char *)&fs+o)+o2)=val;
	if (input_cmds[i].hwrite) input_cmds[i].hwrite(o,o2,input_cmds[i].r);
	return 0;
}

int set_double(int i, char *line)
{
	char *s, *s2;
	double val;
	size_t o;
	int o2;
	o = input_cmds[i].o;
	s=strtok(NULL," ");
	if (s==NULL) return 1;
	o2 = strtol(s,&s2,0);
	if (*s2 != '\0') return 1;  /* offset must not have invalid characters */
	s=strtok(NULL," ");
	val = strtod(s,&s2);
	if (s==s2) return 1;     /* no rval */
	printf("# writing %f to offset 0x%x index %d\n",val,o,o2);
	*((double *)((char *)&fs+o)+o2)=val;
	if (input_cmds[i].hwrite) input_cmds[i].hwrite(o,o2,input_cmds[i].r);
	return 0;
}


int put_int   (FILE *fd, size_t o, int l, char *line);
int put_double(FILE *fd, size_t o, int l, char *line);
int put_short (FILE *fd, size_t o, int l, char *line);

const struct output_cmd {
	char c[2];
	int (*func)(FILE *fd, size_t o, int l, char *line);
	int l;
	size_t o;
} output_cmds[] = {
#define  VAR(t,n,v) { n, put_##t, 1, offsetof(struct fpga_state, v) },
#define IVAR(t,n,v,p,r) { n, put_##t, 1, offsetof(struct fpga_state, v) },
#define  ARRAY_VAR(t,n,v,l) { n, put_##t, l, offsetof(struct fpga_state, v[0]) },
#define ARRAY_IVAR(t,n,v,l,p,r) { n, put_##t, l, offsetof(struct fpga_state, v[0]) },
#include "llrf-state.h"
#undef VAR
#undef IVAR
#undef ARRAY_VAR
#undef ARRAY_IVAR
};

int put_int(FILE *fd, size_t o, int l, char *line)
{
	int i;
	for (i=0; i<l; i++) fprintf(fd, " %d",*((int *)((char *)&fs+o)+i));
	fputc('\n', fd);
	return 0;
}

int put_double(FILE *fd, size_t o, int l, char *line)
{
	int i;
	for (i=0; i<l; i++) fprintf(fd, " %.2f",*((double *)((char *)&fs+o)+i));
	fputc('\n', fd);
	return 0;
}

int put_short(FILE *fd, size_t o, int l, char *line)
{
#ifdef CAN_TOLERATE_SLOW_PRINTF
	int i;
	for (i=0; i<l; i++) fprintf(fd, " %d",*((short *)((char *)&fs+o)+i));
#else
#define CHUNK 32
	int i, j, s;
	char *p, buff[CHUNK*6+2];
	for (i=0; i<l; i+=CHUNK) {
		p=buff+sizeof(buff)-1;
		*p--='\0';
		j = CHUNK;
		if (j+i>l) j=l-i;
		for (j=j-1; j>=0; j--) {
			s=*((short *)((char *)&fs+o)+i+j);
			p=raw_itoa(s,p);
			*p--=' ';
		}
		fputs(p+1,fd);
	}
#undef CHUNK
#endif
	fputc('\n', fd);
	return 0;
}

int output_all(FILE *fd) {
	int i;
	for (i=0; i<sizeof(output_cmds)/sizeof(*output_cmds); i++) {
		fputc(output_cmds[i].c[0], fd);
		fputc(output_cmds[i].c[1], fd);
		output_cmds[i].func(fd, output_cmds[i].o, output_cmds[i].l, NULL);
	}
	fflush(fd);
	return 0;
}

/* from cordic.c */
void FxUnitVec(long *cos, long *sin, long theta);

/* amp between 0.0 and 1.0, phase in degrees */
void fpga_set_iq(int address, double mag, double phase)
{
	long c, s, lp, lm;
	lp = phase*(1<<16);
	if (mag> 1.0) mag= 1.0;
	if (mag<-1.0) mag=-1.0;
	lm = mag*((1<<16)-1);
	FxUnitVec(&c, &s, lp);
	/* printf("FxUnitVec(%ld)=(%ld,%ld)  lm=%ld\n",lp,c,s,lm); */
	outw(address,   (lm*(c>>16))>>15);
	outw(address+4, (lm*(s>>16))>>15);
}

void fpga_set_setpoint(size_t o, int ix, int reg)
{
	fpga_set_iq(reg, fs.set_m*fs.set_s, fs.set_p);
}

void fpga_set_gain(size_t o, int ix, int reg)
{
	fpga_set_iq(reg, fs.gain_m, fs.gain_p);
}

void fpga_set_dac(size_t o, int ix, int reg)
{
	set_dac(ix, fs.extra_dac[ix]);
}

void fpga_set_int(size_t o, int ix, int reg)
{
	outw(reg+ix*4, *((int *)((char *)&fs+o)+ix));
}

int process_command(char *cmd, FILE *out_fd)
{
	char *s;
	int i;
	s = strtok(cmd," ");
	if (s==NULL) return 1;
	if (*s == '#') return 0;   /* comment */
	if (*s == '*') {
		xfer_from_fpga(0);
		output_all(out_fd);
		return 0;
	}
	for (i=0; i<sizeof(input_cmds)/sizeof(*input_cmds); i++) {
		if (*s == input_cmds[i].c[0] && *(s+1) == input_cmds[i].c[1]) {
			input_cmds[i].func(i,cmd);
			return 0;
		}
	}
	printf("# %s: command not found\n",s);
	return 1;  /* command not found */
}

int test_llrf(void)
{
	unsigned r;
	r = inw(0);  if (r != 0x7701) {
		fprintf(stderr,"product code 0x%x not LLRF magic.\n",r);
		exit(2);
	}
	r = inw(4);   /* Firmware version code */

	/* Make sure we can read and write to the feedforward table.
	 * Exercise each data bus bit to 1 and 0.
	 */
	outw(0x3000,0x55); outw(0x3004,0x66);
	if (inw(0x3000) != 0x55 || inw(0x3004) != 0x66) goto fail;
	outw(0x3000,0xAA); outw(0x3004,0x99);
	if (inw(0x3000) != 0xAA || inw(0x3004) != 0x99) goto fail;
	outw(0x3000,0x00); outw(0x3004,0x00);
	if (inw(0x3000) != 0x00 || inw(0x3004) != 0x00) goto fail;

	// sport_init();
	return r;

fail:	fprintf(stderr," address/data path test FAIL\n");
	exit(2);
}

/* Suitable fakery for testing:
 * create a named pipe: mknod llrf_trigger p
 * start this program with -t llrf_trig, then
 * perl -e 'open(T,">llrf_trig")|| die; while (<>) {syswrite T,"\1\0\0\0",4;}'
 * The design usage is with the sns_pulse kernel module attached to
 * the 60 Hz beam trigger pulse.
 */
inline int read_one(int fd)
{
	unsigned int count;
	int s;
	s = read(fd, &count, sizeof(count));
	if (s == 0 || s == -1) {
		perror("read");
		exit(1);
	}
	if (s == sizeof(count)) {
		// printf("%d\n",count);
		return count;
	} else {
		return -1;
	}
}


#if 1   /* XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX */
void usage(const char *a0)
{
	fprintf(stderr,"usage: %s [-p port] [-t trigger] [-m memory]\n",a0);
	fprintf(stderr,"  live:  %s -t /dev/snspulse -p 3030\n",a0);
	fprintf(stderr,"  test:  %s -t trigger_pipe -m fakedata\n",a0);
	exit(1);
}

char *getram_fake(const char *file, size_t length);
FILE *setup_socket(int port);

int main(int argc, char *argv[])
{
#define NFDS 2
	
	int opt, s, fd, nfds;
	int net_port=0;
	char *trigger_f = NULL;
	char *memory_f = NULL;
	struct pollfd ufds[NFDS];
	FILE *net_fd, *in_fd, *out_fd;
	char cmd[100];

	while ((opt = getopt(argc, argv, "p:t:m:")) > 0) {
	    switch (opt) {
		case 'p': net_port  = atoi(optarg);   break;
		case 't': trigger_f = strdup(optarg); break;
		case 'm': memory_f  = strdup(optarg); break;
		default:  usage(argv[0]);
	    }
	}
	if (memory_f) {
		fpga_base = getram_fake(memory_f, 64*1024);
		fs.trace_decimate=3;
		fs.decay_length=100;
		fs.decay_ignore=0;
		fs.pulse_length=4232;
		fs.timer[5]=200;
	} else {
		fprintf(stderr,"going live to FPGA\n");
		fpga_base = getram_big(FPGA_BASE, 64*1024);
		test_llrf();
	}

	if (net_port) {
		net_fd = setup_socket(net_port);
		if (net_fd == NULL) { perror("network"); exit(1); }
		in_fd = net_fd;
		out_fd = net_fd;
	} else {
		in_fd = stdin;
		out_fd = stdout;
	}
	ufds[0].fd = fileno(in_fd);
	ufds[0].events = POLLIN;
	nfds=1;

	if (trigger_f) {
		fd = open(trigger_f,O_RDONLY|O_NONBLOCK);
		if (fd < 0) { perror("open: trigger"); exit(1); }
		// printf("starting main loop, fd=%d\n",fd);
		ufds[1].fd = fd;
		ufds[1].events = POLLIN;
		nfds=2;
	}


	for (;;) {
		s=poll(ufds, nfds, -1);
		if (s == -1) {
			if (errno == EINTR) continue;
			perror("poll");
			exit(1);
		}
		if (ufds[0].revents & POLLIN) {
			if (fgets(cmd, 100, in_fd)) process_command(cmd,out_fd);
			else return 0;
		} else if (nfds==2 && ufds[1].revents & POLLIN) {
			xfer_from_fpga(read_one(ufds[1].fd));
			if (fs.output_enable && fs.cycle_count%fs.output_enable==1) output_all(out_fd);
		}
	}
}
#else
int main(int argc, char *argv[])
{
	char cmd[100];

#ifdef __arm__
	fpga_base = getram_big(FPGA_BASE, 64*1024);
	test_llrf();
#endif
	while (fgets(cmd, 100, stdin)) process_command(cmd);
	return 0;
}
#endif
